//
//     WARNING
//       This code is completely open and available to be modified and changed
//       in any way and by anybody. It is yours once you start to change it.
//       Don't try to patent your derivative code and sue us for infringement.
//       We were here first!
//
//       The reason it is provided this way is for your education and to illustrate some coding
//       ideas. It can cause damage to your Arduino or other hardware if you are not extremely careful.
//       No guarantee is offered and no risk is assumed by the code author.
//       All the risk is on you even if you follow the commments provided in the
//       source code or elsewhere. This is powerful code but you must know it well.
//
//       This code implements the Command Module as imagined by Robot-DHQ. The main purpose of
//       the code is to receive user commands and interact with other Robot-DHQ modules.
//
//       Robot-DHQ is the Robot Design Headquarters
//
//       The human interface paradigm is "immediate response" to any input. No CR or LF is required if
//       the input is a key on a keyboard. This means a key press is similar to a button press.
//       Either is a command that results in an immediate action.
//       The codes does not wait for a button press to be released.
//       Electrically the button press could be the low to high transition or the high to low transition.
//       If the button is behaving "wrongly" one option is to make a change to the code.
//       The user interface includes both menus and "always available" user commands
//
//       Commands produce actions. Actions may start scripts. Scripts consist of commands.
//       One consequence of this is that definite and indefinite loops are possible.
//       Menus and scripts are saved in the TextSource node.
//       Actions primarily involve setting values in nodes which may include moving servos, etc.
//       Even unscripted actions involve more than one simple interaction with a remote node.
//       A set of tools is embedded in the system code to reveal and debug the internal activity of the code.
//
//       Commands are of several types
//         Commands that simply set variables in the command module.
//         Commands that transmit commands to the servo module.
//         Commands that start one or more interactions with the TextSource module.
//         Commands that load a sequence of steps from the TextSorce module.
//         Commands that terminate a running sequence of steps.
//       To view the command interpreter switch logic search for processUserChar() or char_from_UI.
//       To view the console display logic search for consoleRefresh()
//
//       input and output routing
//         the output is serial (to robot and monitor) if the input is keyboard.
//         the output is serial (to robot and monitor) if the input is a button.
//         the output is serial (to monitor) if the input is serial (from robot)
//         the output is LED to indicate command status mode, etc.
//         internal digital settings may indicate an input set on a pot.
//
//       The code in this module has many undocumented features and has not been optimized. You can learn
//       a lot by studying it or you can use it according to the documentation and
//       not read any further. Notes within the source code will give you some ideas
//       for using some of the undocumented features.
//
//       Consider this code a "work in progress". The Command Module is described in
//       this document: D:\Projects\Arduino\Smart Modules\??? Command Module.odt
//
//       This code is designed to run on an Arduino Nano installed on a Robot-DHQ adapter board.
//       The Arduino serial port configured for 115,200 baud accepts commands and displays menus.
//       Check the baud rate first if the menu is not displayed when first connected.
//
//       The commands sent to Robot-DHQ Servo module are sent over the software serial interface.
//       The software serial port has a number of limitations. For example, it is unreliable to have two.
//       It is unreliable at baud rates above 57,600 baud so it is set to 57,600 baud.
//       Also, some delays are required in code to make the output readable.
//
//       The communications with the TextSource node use a bit bang protocol.
//       The limitation of the bit bang protocol is that processing must be 100 percent devoted to
//       the bit bang code when sending or receiving a character.
//       This requires built-in delays between characters limiting total throughput.
//
// how to avoid destroying the Arduino Nano
// 1   the Arduino will not be detected over the USB interface if the Schottky diode
//     has been destroyed. This can happen the +5 volt supply is accidentally
//     shorted to ground. The diode can only handle about 0.5 amps before it blows out.
//     This is therefore the first thing to suspect if the serial connection drops.
//     Therefore, always check for shorts before applying power
//
// 2   input pins are relatively safe to implement as long as they are not
//     connected to a power source with a voltage greater or less than the
//     hardware limitations, usually VCC and GROUND. The S1025 command is
//     provided to produce an input only state.
//

#include <SoftwareSerial.h>
SoftwareSerial gtSerial2(12, 8); // RX is Arduino pin D12, TX is Arduino pin D8

//   timing parameters
//     the timing defaults can be adjusted by changing these parameters here
const byte slow_loop_size = 250;
// implementing a pseudo millisecond timer
const unsigned long fast_loop_size = 1000;

byte rx_byte_from_servo = 0;
char char_from_UI;
bool refresh_UI = 0;
byte sequence_steps_to_do = 0;
byte repeats_to_do = 0;
bool repeating = 0;
byte step_no = 0;
byte verbosity_level = 0;
char cmd_menu = '1';
char seq_menu = '1';
char help_menu = '1'; // 1-5
char parm_type;
byte servo_cmd_mode = 0;
byte command_mode = 0;
// 0 - (C) command menu mode
// 1 - (S) sequence menu mode
// 2 - (A, D, V, W, X, Y, Z) numeric entry mode
// 3 - (?, L, N, O, P) help menu mode
// 4 - (R) history menu command mode
// 5 - (!) sequence manip menu mode
byte saved_mode = 0;
int parm_value_D = 2;
byte parm_value_A = 0;
int parm_value_I = 0;
int parm_value_J = 0;
int parm_value_K = 0;
int parm_value_L = 0;
int parm_value_M = 0;
int parm_value_N = 0;
int parm_value_O = 0;
int parm_value_P = 0;
int parm_value_V = 0;
int parm_value_W = 0;
int parm_value_X = 0;
int parm_value_Y = 0;
int parm_value_Z = 0;
int saved_digital_pin = 2;
int new_value = 0;
int entered_value = 0;
byte digit_count = 0;
byte h_idx = 0;
String myStr = "";
String last_command;
String text_request;
String tmpString;
String menuItem;

const int PROC_ELEMENTS = 14;
char procReturn [PROC_ELEMENTS];
char procPrep [PROC_ELEMENTS];
const int HST_ELEMENTS = 14;
const int HST_SIZE = 12;
char cmdHistory [HST_ELEMENTS] [HST_SIZE];
const int SEQ_ELEMENTS = 16;
const int CMD_SIZE = 12;
int seqDelay [SEQ_ELEMENTS];
char seqCommand [SEQ_ELEMENTS] [CMD_SIZE];
char fromTextModule[121];
byte recentLine = 0;
byte resp_pos = 0;
char fromServoModule[81];
char savResponse[20];
int sourcePin = 0;
int pulseWidth = 0;
byte parmOne = 0;
char parmTwo = 0;
byte send_receive_step = 0;

// indexes and counters
byte slow_counter = 0;
byte seq_step = 0;
byte manip_step = 0;
byte remember_step = 0;
byte last_step = 0;
int seq_pause_counter = 0;

// bit bang vars
const long END_MSG_PERIOD = 90000;
const long LOCKOUT_PERIOD = 20000;
const byte BB_TX_PIN_1 = 9; // bit bang TX is Arduino pin D9
const byte BB_RX_PIN_1 = 11; // bit bang RX is Arduino pin D11
const byte SHORT_PERIOD = 25;
const byte BIT_PERIOD = 50;
const byte LONG_PERIOD = 100;
const byte NEW_CHAR_PERIOD = 200;
const byte START_CHAR_PERIOD = 150;
const int START_MSG_PERIOD = 1000;
unsigned long startMicros = 0;
unsigned long begin_low = 0;
unsigned long begin_high = 0;
unsigned long comm_activity = 0;
char BB_char_in;
bool BB_input_mode = 0;
bool last_BB_RX_PIN_1 = 1;
bool command_input_enabled = 0;
byte chars_in = 0;
byte bit_num = 0;
byte line_action = 0; // no end-of-line action needed when lines of text received from TextSource
byte tx_period_1;
byte tx_period_2;
byte tx_period_3;
byte tx_period_4;
byte tx_period_5;
byte tx_period_6;
byte tx_period_7;

void setup() {
  gtSerial2.begin(57600);  // connection to servo node
  Serial.begin(115200);
  while (!Serial) {
    ; // wait for serial port to connect. Needed for native USB port only
  }

  initializeDigitalArrays();
  Serial.println(F("Command Module"));
  Serial.print(F("END_MSG_PERIOD = "));
  Serial.println(END_MSG_PERIOD);
  Serial.print(F("LOCKOUT_PERIOD = "));
  Serial.println(LOCKOUT_PERIOD);
  Serial.print(F("SHORT_PERIOD = "));
  Serial.println(SHORT_PERIOD);
  Serial.print(F("BIT_PERIOD = "));
  Serial.println(BIT_PERIOD);
  Serial.print(F("LONG_PERIOD = "));
  Serial.println(LONG_PERIOD);
  Serial.print(F("START_CHAR_PERIOD = "));
  Serial.println(START_CHAR_PERIOD);
  Serial.print(F("NEW_CHAR_PERIOD = "));
  Serial.println(NEW_CHAR_PERIOD);
  Serial.print(F("START_MSG_PERIOD = "));
  Serial.println(START_MSG_PERIOD);

  // the next five lines initialize the bit bang code
  BB_input_mode = 1; // bit bang input mode
  begin_high = micros();
  delay(20); // force a timeout
  pinMode(BB_RX_PIN_1, INPUT);
  pinMode(BB_TX_PIN_1, OUTPUT);

  text_request = F("MC1");
  prepOptionMenu();
}

void initializeDigitalArrays() {
  for (h_idx = 0; h_idx < HST_ELEMENTS; h_idx++) {
    myStr = "";
    myStr.toCharArray(cmdHistory[h_idx], HST_SIZE);
  }
  pinMode(LED_BUILTIN, OUTPUT);
  digitalWrite(LED_BUILTIN, LOW);
}

void loop() {
  if (BB_input_mode > 0) {
    // begin bit bang input code
    if (digitalRead(BB_RX_PIN_1) < 1) {
      // BB_RX_PIN_1 low
      if (last_BB_RX_PIN_1 > 0) { // high to low transition
        begin_low = micros();
        if (bit_num < 1) {
          bit_num++; // start of char
        } else {
          if (bit_num < 3) {
            if (begin_low - begin_high > BIT_PERIOD) {
              BB_char_in += 32; // bit 2
            }
            bit_num++;
          } else {
            if (bit_num < 5) {
              if (begin_low - begin_high > BIT_PERIOD) {
                BB_char_in += 8; // bit 4
              }
              bit_num++;
            } else {
              if (bit_num < 7) {
                if (begin_low - begin_high > BIT_PERIOD) {
                  BB_char_in += 2; // bit 6
                }
                bit_num++;
              }
            }
          }
        }
        last_BB_RX_PIN_1 = 0;
      }
    } else {
      // BB_RX_PIN_1 high
      if (last_BB_RX_PIN_1 < 1) { // low to high transition
        begin_high = micros();
        if (bit_num < 2) {
          if (begin_high - begin_low > BIT_PERIOD) {
            BB_char_in = 64; // bit 1
          } else {
            BB_char_in = 0; // bit 1
          }
          bit_num++;
        } else {
          if (bit_num < 4) {
            if (begin_high - begin_low > BIT_PERIOD) {
              BB_char_in += 16; // bit 3
            }
            bit_num++;
          } else {
            if (bit_num < 6) {
              if (begin_high - begin_low > BIT_PERIOD) {
                BB_char_in += 4; // bit 5
              }
              bit_num++;
            } else {
              if (begin_high - begin_low > BIT_PERIOD) {
                BB_char_in += 1; // bit 7
              }
              processBitBangChar();
              bit_num = 0;
            }
          }
        }
        last_BB_RX_PIN_1 = 1;
      }
    }

    // the input signal is high when nothing is being transmitted
    // detect this and drop out of bit bang input mode
    // also set bit_num to zero to "synchronize" the interface
    // end message period must be long enough to cover any
    // "jitter" in the incoming bit stream
    if (micros() - begin_high > END_MSG_PERIOD) {
      BB_input_mode = 0;
      bit_num = 0;
    }


    // end bit bang input code
  } else {
    // begin not bit bang input code
    if (micros() - startMicros >= fast_loop_size) {
      // check clock and send queued and scheduled commands to remote processor.
      slowLoop();          // once per millisecond, 1000 times per second, period = 1000 microseconds
      startMicros = micros();
    }

    while (Serial.available() > 0) { // command interface
      char_from_UI = Serial.read();
      if (micros() - comm_activity > LOCKOUT_PERIOD) { // period of comm inactivity > lockout period
        command_input_enabled = 1; // unlock command console
      }
      if (command_input_enabled > 0) {
        processUserChar();
      } else {
        commandInputDisabled();
      }
    }

    if (gtSerial2.available()) { // servo node
      rx_byte_from_servo = gtSerial2.read();
      processServoChar();
    }
    // end not bit bang input code
  }
}

void processBitBangChar() {
  if (4 == BB_char_in) { // ASCII EOT
    BB_input_mode = 0; // allow processing of non-BB serial input
    processEOM();
  } else {
    if (13 == BB_char_in) { // ASCII CR
      fromTextModule[chars_in] = '\0';
      comm_activity = micros();
      end_of_line_recd();
      chars_in = 0;
    } else {
      fromTextModule[chars_in] = BB_char_in;
      if (chars_in < 120) {
        chars_in++;
      }
    }
    begin_high = micros();
  }
}

void processEOM() {
  if (send_receive_step > PROC_ELEMENTS - 2) {
    Serial.println();
    Serial.print(F("PROC error!")); // need to increase the size of PROC_ELEMENTS
  }
  comm_activity = micros();
  if (verbosity_level > 1) {
    Serial.print(F("<r"));
    Serial.print(procReturn [send_receive_step]);
    Serial.print(F(">"));
  }
  switch (procReturn [send_receive_step]) {
    case 'a': // print command description text from TextSource node
      Serial.println();
      Serial.print(line_key(recentLine));
      Serial.print(F(" - "));
      Serial.print(fromTextModule);
      break;
    case 'b': // print limits report
      limitReport(); // tag this on the end of H05
      command_input_enabled = 1;
      break;
    case 'c': // print variables report
      varReport(); // tag this on the end of H05
      command_input_enabled = 1;
      break;
    case 'd': // print the command description from the TextSource node
      Serial.println();
      Serial.println();
      Serial.print(fromTextModule);
      command_input_enabled = 1;
      break;
    case 'e': // manipulate procReturn and procPrep
      procManip();
      break;
  }
  requestDataFromRemoteNode(); // get the servo command or sequence from the TextSource node
}

void procManip() {
  if (verbosity_level > 1) {
    Serial.print(F("pr:"));
  }
  for (int i = 0; i < PROC_ELEMENTS; i++) {
    if (i > 0) {
      if (i < ((String) fromTextModule).length() + 1) {
        procPrep [i] = fromTextModule[i - 1];
      } else {
        if (i < ((String) fromTextModule).length() + 2) {
          procPrep [i] = '2';
        }
      }
    }
    if (verbosity_level > 1) {
      if (procPrep [i] > '0' - 1) {
        Serial.print(procPrep [i]);
      }
    }
  }
  command_input_enabled = 1;
}

void slowLoop() { // millisecond
  slow_counter++;
  if (slow_counter > slow_loop_size) {
    slow_counter = 0;
    // these actions are done once per 250 millisecond, 4 times per second, period = 250,000 microseconds
    toggleLED();
  }
  if (sequence_steps_to_do > 0) {
    doStoredStep();
  }
}

// this function includes switch debounce logic
void toggleLED() {
  if (digitalRead(LED_BUILTIN) < 1) {
    digitalWrite(LED_BUILTIN, HIGH); // LED_BUILTIN is 13
  } else {
    digitalWrite(LED_BUILTIN, LOW);
  }
}

// output of the servo node processor is buffered and displayed when a CR is received.
void processServoChar() {
  comm_activity = micros();
  if (rx_byte_from_servo != 10) { // newline
    if (rx_byte_from_servo > 31) {
      if (rx_byte_from_servo < 127) {
        if (resp_pos < 80) {
          fromServoModule[resp_pos] = rx_byte_from_servo;
          resp_pos++;
        }
      }
    }
  }
  if (rx_byte_from_servo == 13) {
    fromServoModule[resp_pos] = '\0';
    resp_pos = 0;
    processServoResponse();
  }
}

void processServoResponse() {
  if (servo_cmd_mode < 1) {
    Serial.println();
    Serial.print(last_command);
    Serial.print(F(": "));
    Serial.print(fromServoModule);
  }
  entered_value = numericResponse(last_command, 2, 5);
  // the entered value in this case is the three-digit command
  switch (entered_value) {
    case 965: // pulse loop size
      setParm('V', numericResponse(fromServoModule, 4, 5));
      break;
    case 975: // slew rate
      setParm('W', numericResponse(fromServoModule, 5, 5));
      break;
    case 977: // max pulse width
      setParm('Z', numericResponse(fromServoModule, 0, 5));
      switch (parm_value_D) { // current digital pin #
        case 2:
          parm_value_M = parm_value_Z;
          break;
        case 3:
          parm_value_N = parm_value_Z;
          break;
        case 4:
          parm_value_O = parm_value_Z;
          break;
        case 5:
          parm_value_P = parm_value_Z;
          break;
      }
      break;
    case 978: // min pulse width
      setParm('Y', numericResponse(fromServoModule, 0, 5));
      switch (parm_value_D) { // current digital pin #
        case 2:
          parm_value_I = parm_value_Y;
          break;
        case 3:
          parm_value_J = parm_value_Y;
          break;
        case 4:
          parm_value_K = parm_value_Y;
          break;
        case 5:
          parm_value_L = parm_value_Y;
          break;
      }
      break;
    case 979: // target pulse width
      setParm('X', numericResponse(fromServoModule, 0, 5));
      break;
    case 980: //
      entered_value = numericResponse(fromServoModule, 0, 5);
      pulseWidth = entered_value;
      break;
    case 984: //
      entered_value = numericResponse(fromServoModule, 0, 5);
      sourcePin = entered_value;
      break;
    case 985: //
      savResponse[11] = fromServoModule [0];
      savResponse[12] = fromServoModule [1];
      savResponse[13] = fromServoModule [2];
      savResponse[14] = fromServoModule [3];
      savResponse[15] = fromServoModule [4];
      savResponse[16] = fromServoModule [5];
      savResponse[17] = fromServoModule [6];
      savResponse[18] = fromServoModule [7];
      break;
    case 986: //
      savResponse[8] = fromServoModule [0];
      savResponse[9] = fromServoModule [1];
      savResponse[10] = fromServoModule [2];
      break;
    case 987: //
      savResponse[2] = fromServoModule [0];
      savResponse[3] = fromServoModule [1];
      savResponse[4] = fromServoModule [2];
      break;
    case 988: //
      savResponse[5] = fromServoModule [0];
      savResponse[6] = fromServoModule [1];
      savResponse[7] = fromServoModule [2];
      break;
    case 989: // current analog pin
      parm_value_A = numericResponse(fromServoModule, 1, 3);
      if (servo_cmd_mode > 0) {
        Serial.println();
        Serial.print(F("A"));
        prtAnalogPin();
        servo_cmd_mode = 0;
      }
      break;
    case 990: // current digital pin
      parm_value_D = numericResponse(fromServoModule, 1, 3);
      if (servo_cmd_mode > 0) {
        Serial.println();
        Serial.print(F("D"));
        prtDigitalPin();
        servo_cmd_mode = 0;
      }
      break;
    default:
      if (entered_value > 99) {
        if (entered_value < 990) {
          Serial.println();
          Serial.print(F("command = "));
          Serial.print(last_command);
          Serial.print(F(", response = "));
          Serial.print(fromServoModule);
          Serial.println(F(", data was not saved."));
        }
      }
      break;
  }
  command_input_enabled = 1;
}

void commandInputDisabled() {
  switch (char_from_UI) {
    case 10: // newline (ignored)
      break;
    case 13: // CR (ignored)
      break;
    default:
      Serial.println();
      Serial.print(F("a command ("));
      Serial.print(char_from_UI);
      Serial.print(F(") was ignored."));
      break;
  }
}

// each character is processed as an independent command as it is entered
// this processing varies according to the current console mode.
// +, -, . and CR terminate numeric entry and affect a value.
// other characters terminate numeric entry and leave the value unchanged.
void processUserChar() {
  if (char_from_UI > 96) {
    if (char_from_UI < 123) {
      char_from_UI -= 32; // treat upper and lower case the same
    }
  }
  switch (char_from_UI) {
    case 3: // control-C
      terminateCurrentSeq();
      break;
    case 10: // newline (ignored)
      break;
    case 13: // carriage return
      captureValue();
      break;
    case 33: // ! - sequence debug
      command_mode = 5; // sequence manip menu mode
      refresh_UI = 1;
      break;
    case 43: // +
      processIncrementKey();
      break;
    case 45: // -
      processDecrementKey();
      break;
    case 46: // . (period)
      captureValue();
      break;
    case 47: // / (forward slash)
      VaryVerbosity();
      break;
    case 48:
    case 49:
    case 50:
    case 51:
    case 52:
    case 53:
    case 54:
    case 55:
    case 56:
    case 57:
      processDigitKey();
      break;
    case 63: // ? main help page
      saved_mode = command_mode;
      command_mode = 3; // help menu mode
      help_menu = '1'; // H01
      refresh_UI = 1;
      break;
    case 65: // A
      enterParm('A');
      break;
    case 67: // C
      if (0 == command_mode) {
        saved_mode = command_mode;
        command_mode = 2; // numeric input mode
        parm_type = 'C'; // command menu number
      } else {
        command_mode = 0; // (C) command menu mode
      }
      refresh_UI = 1;
      break;
    case 68: // D
      enterParm('D');
      break;
    case 69: // E - menu option key
    case 70: // F - menu option key
    case 71: // G - menu option key
    case 72: // H - menu option key
      doMenuSelection();
      break;
    case 76: // L - display the servo limits help page
      saved_mode = command_mode;
      command_mode = 3; // help menu mode
      help_menu = '5'; // H05
      refresh_UI = 1;
      break;
    case 77: // M - re-display the current menu
      if (2 == command_mode) {
        command_mode = saved_mode;
      }
      refresh_UI = 1;
      break;
    case 78: // N - display the number entry help page
      saved_mode = command_mode;
      command_mode = 3; // help menu mode
      help_menu = '4'; // H04
      refresh_UI = 1;
      break;
    case 79: // O - display the operations help page
      saved_mode = command_mode;
      command_mode = 3; // help menu mode
      help_menu = '3'; // H03
      refresh_UI = 1;
      break;
    case 80: // P - display the parm values help page
      saved_mode = command_mode;
      command_mode = 3; // help menu mode
      help_menu = '2'; // H02
      refresh_UI = 1;
      break;
    case 81: // Q - quit sequencing
      terminateCurrentSeq();
      break;
    case 82: // R - recent commands
      command_mode = 4; // recent command mode
      refresh_UI = 1;
      break;
    case 83: // S
      if (1 == command_mode) {
        saved_mode = command_mode;
        command_mode = 2; // numeric input mode
        parm_type = 'S'; // sequence menu
      } else {
        command_mode = 1; // (S) sequence menu mode
      }
      refresh_UI = 1;
      break;
    case 84: // T
      VaryVerbosity();
      break;
    case 86: // V
    case 87: // W
    case 88: // X
    case 89: // Y
    case 90: // Z
      enterParm(char_from_UI);
      break;
    default:
      Serial.println();
      Serial.println();
      Serial.print(F("No command has been coded for this key press: "));
      Serial.print(char_from_UI);
      Serial.print(F(" ("));
      Serial.print((byte)char_from_UI);
      Serial.println(F(") "));
      break;
  }
  consoleRefresh(); // the "console" includes menus, prompts and help displays
}

void enterParm(char the_parm) {
  digit_count = 0;
  saved_mode = command_mode;
  command_mode = 2; // numeric input mode
  parm_type = the_parm;
  refresh_UI = 1;
}

void processIncrementKey() {
  switch (command_mode) {
    case 0: // (C) command menu mode
      cmd_menu++;
      if (cmd_menu > '8') {
        cmd_menu = '8';
        seq_menu = '1';
        command_mode = 1; // (S) sequence menu mode
      }
      refresh_UI = 1;
      break;
    case 1: // (S) sequence menu mode
      seq_menu++;
      if (seq_menu > '2') {
        seq_menu = '2';
        command_mode = 4; // recent command mode
      }
      refresh_UI = 1;
      break;
    case 2: // numeric input mode
      plusMinusValue(1);
      break;
    case 4: // recent command mode
      command_mode = 5; // sequence manip menu mode
      refresh_UI = 1;
      break;
    case 5: // sequence manip menu mode
      command_mode = 0; // (C) command menu mode
      cmd_menu = '1';
      refresh_UI = 1;
      break;
  }
}

void processDecrementKey() {
  switch (command_mode) {
    case 0: // (C) command menu mode
      if (cmd_menu > '1') {
        cmd_menu--;
      } else {
        command_mode = 5; // sequence manip menu mode
      }
      refresh_UI = 1;
      break;
    case 1: // (S) sequence menu mode
      seq_menu--;
      if (seq_menu < '1') {
        seq_menu = '1';
        cmd_menu = '8';
        command_mode = 0; // (C) command menu mode
      }
      refresh_UI = 1;
      break;
    case 2: // numeric input mode
      plusMinusValue(-1);
      break;
    case 4: // recent command mode
      command_mode = 1; // (S) sequence menu mode
      seq_menu = '2';
      refresh_UI = 1;
      break;
    case 5: // sequence manip menu mode
      command_mode = 4; // recent command mode
      refresh_UI = 1;
      break;
  }
}

void plusMinusValue(int adjustment) {
  if ('A' == parm_type) {
    changeAnalogPinServo(parm_value_A + adjustment);
  } else if ('D' == parm_type) {
    changeDigitalPinServo(parm_value_D + adjustment);
  } else {
    setParm(parm_type, limited_value(getParm(parm_type) + adjustment, 0, 32766));
  }
  command_mode = saved_mode;
}

void processDigitKey() {
  switch (command_mode) {
    case 0: // (C) command menu mode
    case 1: // (S) sequence menu mode
      doMenuSelection();
      break;
    case 2: // numeric input mode
      digitEntered();
      break;
    default: // history and manip modes
      doMenuSelection();
      break;
  }
}

void digitEntered() {
  switch (parm_type) {
    case 'C':
      cmd_menu = char_from_UI;
      if (cmd_menu > '8') {
        cmd_menu = '8';
      }
      if (cmd_menu < '1') {
        cmd_menu = '1';
      }
      command_mode = 0; // (C) command menu mode
      break;
    case 'S':
      seq_menu = char_from_UI;
      if (seq_menu > '2') {
        seq_menu = '2';
      }
      if (seq_menu < '1') {
        seq_menu = '1';
      }
      command_mode = 1; // (S) sequence menu mode
      break;
    default:
      Serial.print(char_from_UI);
      if (digit_count < 1) {
        entered_value = 0;
      }
      entered_value = entered_value * 10 + (char_from_UI - 48);
      digit_count++;
      break;
  }
  refresh_UI = 1;
}

void captureValue() {
  if (digit_count > 0) {
    if ('A' == parm_type) {
      changeAnalogPinServo(entered_value);
    } else if ('D' == parm_type) {
      changeDigitalPinServo(entered_value);
    } else {
      setParm(parm_type, limited_value(entered_value, 0, 32766));
    }
    command_mode = saved_mode;
  }
}

void consoleRefresh() {
  if (refresh_UI > 0) {
    refresh_UI = 0;
    switch (command_mode) {
      case 0: // (C) command menu mode
        text_request = F("MC");
        text_request += cmd_menu;
        prepOptionMenu();
        break;
      case 1: // (S) sequence menu mode
        text_request = F("MS");
        text_request += seq_menu;
        prepOptionMenu();
        break;
      case 2: // numeric input mode
        valueEntryConsole();
        break;
      case 3: // help menu mode
        prepHelpText();
        break;
      case 4: // recent command mode
        displayRecent();
        break;
      case 5: // sequence manip menu mode
        stepListMenu();
        break;
    }
  }
}

void prepHelpText() {
  command_mode = saved_mode;
  text_request = F("H0");
  text_request += help_menu;
  switch (help_menu) {
    case '4':
      beginRequestProcess(F(" c"), F("2"));
      break;
    case '5':
      beginRequestProcess(F(" b"), F("2"));
      break;
    default:
      beginRequestProcess(F(""), F("2"));
      break;
  }
}

void prepOptionMenu() {
  beginRequestProcess(F(" e"), F("E"));
}

void valueEntryConsole() {
  switch (parm_type) {
    case 'C':
      Serial.println();
      Serial.println(F("Enter command menu #: "));
      break;
    case 'S':
      Serial.println();
      Serial.println(F("Enter sequence menu #: "));
      break;
    case 'A':
      valueEntryMode(parm_value_A, F("analog pin"));
      break;
    case 'D':
      valueEntryMode(parm_value_D, F("digital pin"));
      break;
    case 'V':
    case 'W':
    case 'X':
    case 'Y':
    case 'Z':
      parmEntryMode(parm_type);
      break;
  }
}

void parmEntryMode(char parm_char) {
  valueEntryMode(getParm(parm_char), String("parm ") + parm_char);
}

void displayRecent() {
  Serial.println();
  Serial.println();
  Serial.println(F("Recent Commands"));
  h_idx = 0;
  while (h_idx < HST_ELEMENTS) {
    if (((String)cmdHistory[h_idx]).length() > 0) {
      recentLine = h_idx;
      h_idx = HST_ELEMENTS;
      beginRequestProcess(F(" a"), F("G1"));
    } else {
      Serial.println();
      Serial.print(line_key(h_idx));
      Serial.print(F(" - "));
      h_idx++;
    }
  }
}

void stepListMenu() {
  Serial.println();
  Serial.println();
  Serial.print(F("1 - List all, 2 - Back, 3 - Next, 4 - Single Step, 5 - Run"));
}

void varReport() {
  Serial.println();
  Serial.print(F("Current a"));
  prtAnalogPin();
  Serial.println();
  Serial.print(F("Current d"));
  prtDigitalPin();
  prtParmValue('V');
  prtParmValue('W');
  prtParmValue('X');
  prtParmValue('Y');
  prtParmValue('Z');
}

void prtAnalogPin() {
  Serial.print(F("nalog pin is A"));
  Serial.print(parm_value_A);
}

void prtDigitalPin() {
  Serial.print(F("igital pin is D"));
  Serial.print(parm_value_D);
}

void prtParmValue(char label) {
  Serial.println();
  Serial.print(F("Current value of parm "));
  Serial.print(label);
  Serial.print(F(" is "));
  Serial.print(getParm(label));
}

void valueEntryMode(int current_value, String label) {
  if (digit_count < 1) {
    Serial.println();
    Serial.println();
    Serial.print(F("The current value for "));
    Serial.print(label);
    Serial.print(F(" is "));
    Serial.println(current_value);
    Serial.println(F("Enter a new value followed by . (period) or return."));
  }
}

// a menu item will be described and executed
void doMenuSelection() {
  switch (command_mode) {
    case 0: // (C) command menu mode
      menuItem = F("DC");
      menuItem += cmd_menu;
      menuItem += char_from_UI;
      prepMenuItem();
      break;
    case 1: // (S) sequence menu mode
      menuItem = F("DS");
      menuItem += seq_menu;
      menuItem += char_from_UI;
      prepMenuItem();
      break;
    case 4: // recent command mode
      Serial.println();
      if (strlen(cmdHistory[line_index(char_from_UI)]) > 0) {
        menuItem = cmdHistory[line_index(char_from_UI)];
        prepMenuItem();
      } else {
        Serial.print(F("That is not a valid command."));
      }
      break;
    case 5: // sequence manip menu mode
      doStepListOption();
      break;
  }
}

// manipulate the step list loaded in memory
void doStepListOption() {
  refresh_UI = 1;
  switch (char_from_UI) {
    case '1':
      printAllSteps();
      break;
    case '2':
      if (manip_step > 0) {
        manip_step--;
      }
      refresh_UI = 0;
      printStep(manip_step);
      break;
    case '3':
      if (manip_step < last_step) {
        manip_step++;
      }
      printStep(manip_step);
      refresh_UI = 0;
      break;
    case '4':
      doSingleStep();
      break;
    case '5': // run
      seq_step = manip_step;
      sequence_steps_to_do = last_step - seq_step + 1;
      refresh_UI = 0;
      break;
    default:
      Serial.println();
      Serial.print(F("Undefined: "));
      Serial.print(char_from_UI);
      break;
  }
}

void doSingleStep() {
  if (((String)seqCommand[manip_step]).length() > 0) {
    seq_step = manip_step;
    sequence_steps_to_do = 1;
    if (manip_step < last_step) {
      manip_step++;
      refresh_UI = 0;
    }
    if (manip_step < last_step) {
      printStep(manip_step);
    }
  } else {
    Serial.println();
    Serial.print(F("Step is empty"));
  }
  //  printStepWithHeader();
}

void printStepWithHeader() {
  Serial.println();
  Serial.println();
  if (((String)seqCommand[manip_step]).length() > 0) {
    Serial.println(F("  Delay  Command"));
    printStep(manip_step);
  } else {
    Serial.println(F("Step is empty"));
  }
}

void printAllSteps() {
  Serial.println();
  Serial.println();
  if (((String)seqCommand[0]).length() > 0) {
    Serial.println(F("List content:"));
    Serial.println(F("  Delay  Command"));
  } else {
    Serial.print(F("List is empty"));
  }
  h_idx = 0;
  while (h_idx < SEQ_ELEMENTS) {
    if (((String)seqCommand[h_idx]).length() > 0) {
      printStep(h_idx);
      last_step = h_idx;
    }
    h_idx++;
  }
}

void printStep(byte element_idx) {
  Serial.println();
  if (element_idx != manip_step) {
    Serial.print(F("  "));
  } else {
    Serial.print(F("* "));
  }
  with_prefix_spaces(seqDelay[element_idx]);
  Serial.print(F("  "));
  Serial.print(seqCommand[element_idx]);
}

// the number returned is from 0 to 13
char line_index(char char_entered) {
  if (char_entered < 49) {
    return 9;
  } else {
    if (char_entered < 58) {
      return char_entered - 49;
    } else {
      return char_entered - 59;
    }
  }
}

void with_prefix_spaces(int num_to_prefix) {
  if (num_to_prefix < 100) {
    Serial.print(F(" "));
  }
  if (num_to_prefix < 1000) {
    Serial.print(F(" "));
  }
  if (num_to_prefix < 10000) {
    Serial.print(F(" "));
  }
  Serial.print(num_to_prefix);
}

char line_key(char line_no) {
  if (line_no < 9) {
    return line_no + 49;
  } else {
    if (line_no < 10) {
      return '0';
    } else {
      return line_no + 59;
    }
  }
}

void terminateCurrentSeq() {
  if (sequence_steps_to_do > 0) {
    sequence_steps_to_do = 0;
    Serial.println();
    Serial.println(F("Sequence terminated"));
    refresh_UI = 1;
  }
}

void end_of_line_recd() {
  comm_activity = micros();
  myStr = (String)fromTextModule;
  if (verbosity_level > 3) {
    Serial.println();
    Serial.print(F("Line From Text Source: "));
    Serial.print(myStr);
  }
  switch (line_action) {
    case 1: // display the text
      if (myStr.length() > 73) {        // split long lines
        if (myStr.indexOf("~") > 1) {
          Serial.println();
          Serial.println(myStr.substring(0, myStr.indexOf("~") - 1));
          Serial.print(F("          "));
          Serial.print(((String)fromTextModule).substring(myStr.indexOf("~")));
        } else {
          Serial.println();
          Serial.println(myStr.substring(0, myStr.indexOf("(") - 1));
          Serial.print(F("          "));
          Serial.print(((String)fromTextModule).substring(myStr.indexOf("(")));
        }
      } else {
        Serial.println();
        Serial.print(fromTextModule);
      }
    case 2: // text is a command (or step in a sequence)
      if (myStr.startsWith("step")) {
        set_step(myStr.substring(4, myStr.indexOf(',')).toInt(),
                 myStr.substring(myStr.indexOf(',') + 1));
      }
      break;
  }
}

// this function merely stores the steps in an array
// for more details look at processScriptStep
void set_step(int cmd_delay, const String & cmd_str) {
  if (step_no < SEQ_ELEMENTS) {
    if (step_no < 1) {
      for (h_idx = 0; h_idx < SEQ_ELEMENTS; h_idx++) {
        myStr = "";
        myStr.toCharArray(seqCommand[h_idx], CMD_SIZE);
        seqDelay[h_idx] = 0;
      }
    }
    cmd_str.toCharArray(seqCommand[step_no], CMD_SIZE);
    seqDelay[step_no] = cmd_delay;
    step_no++;
    sequence_steps_to_do = step_no;
    seq_step = 0;
  } else {
    Serial.println(F("It seems there are too many steps defined for this sequence"));
  }
}

void doStoredStep() {
  seq_pause_counter++;
  if (seq_pause_counter > seqDelay[seq_step]) { // pause before sending the next command
    seq_pause_counter = 0;
    if (verbosity_level > 2) {
      Serial.println();
      Serial.print(F("doScriptStep "));
      Serial.print(seq_step);
      Serial.print(F(" ("));
      Serial.print(seqCommand[seq_step]);
      Serial.print(F("); to do = "));
      Serial.print(sequence_steps_to_do);
    }
    remember_step = seq_step;
    processScriptStep(seqCommand[seq_step]);
    if (remember_step > seq_step) { // the 'R' code causes a decrease in seq_step
      processScriptStep(seqCommand[seq_step]);
    }
    seq_step++;
    sequence_steps_to_do--;
  }
}


// examine this code to understand how to code script steps
// the first letter in the step is the key to what it does
void processScriptStep(String command_string) {
  switch (command_string.charAt(0)) {
    case 'A': // assign value to a parm
      setParm(command_string.charAt(1), numericResponse(command_string, 2, 5));
      break;
    case 'M': // execute a Menu option
      if (verbosity_level > 3) {
        menuItem = "D";
        menuItem += command_string.substring(1, 4);
        prepMenuItem();
      } else {
        menuItem = "C";
        menuItem += command_string.substring(1, 4);
        beginRequestProcess(F("d "), F("C9"));
      }
      break;
    case '>': // remember the current digital pin
      saved_digital_pin = parm_value_D;
      break;
    case '<': // restore the current digital pin
      changeDigitalPinServo(saved_digital_pin);
      refresh_UI = 0; // reverse what limited_value did, i.e. don't refresh menu
      break;
    case 'P': // print report
      Serial.println();
      Serial.println();
      switch (command_string.charAt(1) - 48) {
        case 1:
          report_one();
          break;
        case 2:
          report_two();
          break;
      }
      break;
    case 'R': // return x steps and repeat y times
      parmOne = command_string.charAt(1) - 48;
      parmOne = parmOne * 10 + command_string.charAt(2) - 48;
      parmTwo = command_string.charAt(3) - 48;
      parmTwo = parmTwo * 10 + command_string.charAt(4) - 48;
      if (parmTwo < 1) {
        repeats_to_do = 1; // will repeat indefinitely
      } else {
        if (repeating > 0) {
          if (repeats_to_do < 1) {
            repeating = 0;
            sequence_steps_to_do = 0;
            refresh_UI = 1;
          }
        } else {
          repeats_to_do = parmTwo;
          repeating = 1;
        }
      }
      if (repeats_to_do > 0) {
        parmOne++;
        sequence_steps_to_do += parmOne;
        seq_step -= parmOne;
        repeats_to_do--;
      }
      break;
    case 'S':
      sendToServo(command_string);
      break;
    default:
      Serial.println();
      Serial.print(F("? ")); // not a valid sequence command
      Serial.println(command_string);
      break;
  }
}

// the description and command(s) come from the TextSource node
// first use menuItem to request and print the description of the selection
// then modify menuItem to get the command or steps to be done
void prepMenuItem() {
  beginRequestProcess(F(" d "), F("FC9"));
}

int numericResponse(String src_string, int start_pos, int num_digits) {
  return src_string.substring(start_pos, start_pos + num_digits).toInt();
}

// prep values
// 1 = get description of a recent command for menu
// 2 = request menu or help text
// 3 = set parm Z
// 4 = set parm Y
// 5 = set parm X
// 6 = set parm W
// 7 = set parm V
// 8 = set digital pin (SD)
// 9 = send servo command
// A = get servo command(s) from TextSource
// B = set analog pin (SA)
// return values
//   = silent / do nothing
// a = displaying recent commands menu
// b = print limits report
// c = print varbles report
// d = print the command description
void beginRequestProcess(String return_values, String prep_values) {
  for (int i = 0; i < PROC_ELEMENTS; i++) {
    procPrep [i] = 0;
    if (i < prep_values.length()) {
      procPrep [i] = prep_values.charAt(i);
    }
    procReturn [i] = 0;
    if (i < return_values.length()) {
      procReturn [i] = return_values.charAt(i);
    }
  }
  send_receive_step = 0;
  requestDataFromRemoteNode();
}

void requestDataFromRemoteNode() {
  if (verbosity_level > 1) {
    Serial.print(F("<p"));
    Serial.print(procPrep [send_receive_step]);
    Serial.print(F(">"));
  }
  line_action = 0; // no end-of-line action needed when lines of text received from TextSource
  switch (procPrep [send_receive_step]) {
    case '1': // request command description
      if (recentLine < HST_ELEMENTS) {
        recentLine++;
        sendToTextSource(cmdHistory[recentLine]);
        send_receive_step = -1; // repeat console step
      }
      break;
    case '2': // request menu or help text
      Serial.println();
      Serial.println();
      line_action = 1; // display multiple lines of text
      sendToTextSource(text_request); // request menu or help text
      break;
    case '9': // send a command to the servo node
      myStr = (String)fromTextModule;
      if (myStr.startsWith("S1")) { // servo command
        if (myStr.startsWith("S1003,")) { // set digital pin
          changeDigitalPinServo(myStr.substring(6).toInt());
        } else if (myStr.startsWith("S1004,")) { // set analog pin
          changeAnalogPinServo(myStr.substring(6).toInt());
        } else {
          sendToServo(myStr);
        }
      } else {
        Serial.println();
        Serial.print(myStr);
      }
      break;
    case 'A': // set analog pin in the TextSource node
      setTextSourceParmValue(procPrep [send_receive_step]);
      break;
    case 'C': // get servo command or commands
      menuItem.setCharAt(0, 'C'); // changes Dxx to Cxx
      line_action = 2; // line is a command or step
      step_no = 0; // prepare to store steps
      sendToTextSource(menuItem); // get the servo command or sequence from the TextSource node
      break;
    case 'D': // set parm_value_D in TextSource node
      setTextSourceParmValue(procPrep [send_receive_step]);
      break;
    case 'E': // get the parms needed by the menu builder
      myStr = text_request;
      myStr.setCharAt(0, 'P'); // changes MC1 to PC1, for example
      sendToTextSource(myStr);
      break;
    case 'F': // request a menu item description
      saveRecent(menuItem);
      sendToTextSource(menuItem);
      break;
    case 'G': // build the menu of recent commands
      sendToTextSource(cmdHistory[recentLine]);
      break;
    case 'I': // set parm I in TextSource node
    case 'J': // set parm J in TextSource node
    case 'K': // set parm K in TextSource node
    case 'L': // set parm L in TextSource node
    case 'M': // set parm M in TextSource node
    case 'N': // set parm N in TextSource node
    case 'O': // set parm O in TextSource node
    case 'P': // set parm P in TextSource node
    case 'V': // set parm V in TextSource node
    case 'W': // set parm W in TextSource node
    case 'X': // set parm X in TextSource node
    case 'Y': // set parm Y in TextSource node
    case 'Z': // set parm Z in TextSource node
      setTextSourceParmValue(procPrep [send_receive_step]);
      break;
  }
  send_receive_step++;
}

// set a value in the TextSource node
void setTextSourceParmValue(char parm_char) {
  tmpString = F("S");
  tmpString += parm_char;
  tmpString += getParm(parm_char);
  sendToTextSource(tmpString);
}

// reading the analog pin value in the servo node sets the analog pin parm when the command node restarts
// the values in the two nodes should always match after that
// analog pin can change by reading the value in the servo node, by a user command or by a step in a script
// reading the value in the servo node (the S0989 command) sets the analog pin parm in the command node
// the user commands, A, + and - set the analog pin parm in the command node
// the S1004 command changes the analog pin in the servo node
// when a user command changes the analog pin parm a S1004 command must be sent
// when a script step loaded from the TextSource node changes the analog pin a S1004 command must be sent
// the TextSource node does not keep up with parm_value_A and needs to be informed before it displays any menu containing it
void changeAnalogPinServo(byte proposed_value) {
  new_value = limited_value(proposed_value, 0, 7);
  if (parm_value_A != new_value) {
    parm_value_A = new_value;
    tmpString = F("S1004,");
    tmpString += twoDigitParm(parm_value_A);
    sendToServo(tmpString);
  }
}

// digital pin can change by reading the value in the servo node, by a user command or by a TextSource command
// reading the value in the servo node is the S0990 command
// the user commands, D, + and - are processed directly by the command node user interface
// the TextSource changes the digital pin when the command is the S1003 command
// reading the value in the servo node sets parm_value_D whenever command restarts and the values should always match after that
// the TextSource node does not keep up with parm_value_D and needs to be informed before it displays any menu containing it
// reading the value in the servo node sets parm_value_D in the command node
// a change by user command sets parm_value_D in the servo node
// a change by a TextSource command sets parm_value_D in the command node
// before displaying a menu that displays parm_value_D it is set in the TextSource node by the command node
void changeDigitalPinServo(int proposed_value) {
  new_value = limited_value(proposed_value, 2, 13);
  if (parm_value_D != new_value) {
    parm_value_D = new_value;
    tmpString = F("S1003,");
    tmpString += twoDigitParm(parm_value_D);
    sendToServo(tmpString);
  }
}

int limited_value(int new_value, int lo_value, int hi_value) {
  if (new_value < lo_value) {
    new_value = lo_value;
  }
  if (new_value > hi_value) {
    new_value = hi_value;
  }
  refresh_UI = 1;
  return new_value;
}

void report_one() {
  Serial.println(F("Report 0ne"));
  Serial.println();
  Serial.print(F("Pin D"));
  Serial.print(parm_value_D);
  Serial.print(F(", Function: "));
  Serial.print(savResponse[2]);
  Serial.print(savResponse[3]);
  Serial.print(savResponse[4]);
  Serial.print(F(", SWI/LOC: "));
  Serial.print(savResponse[5]);
  Serial.print(savResponse[6]);
  Serial.print(savResponse[7]);
  Serial.print(F(", Direction: "));
  Serial.print(savResponse[8]);
  Serial.print(savResponse[9]);
  Serial.println(savResponse[10]);
  Serial.print(F("Control: "));
  Serial.print(savResponse[11]);
  Serial.print(savResponse[12]);
  Serial.print(savResponse[13]);
  Serial.print(savResponse[14]);
  Serial.print(savResponse[15]);
  Serial.print(savResponse[16]);
  Serial.print(savResponse[17]);
  Serial.print(savResponse[18]);
  Serial.print(F(", Source Pin: "));
  Serial.println(sourcePin);
  Serial.print(F("Current Pulse Width: "));
  Serial.print(pulseWidth);
  Serial.print(F(", Target Pulse Width: "));
  Serial.println(parm_value_X);
  Serial.print(F("Slew Rate "));
  Serial.println(parm_value_W);
}

void report_two() {
  Serial.println(F("Report Two"));
  Serial.println();
  limitReport();
  Serial.println();
}

void limitReport() {
  Serial.println();
  Serial.print(F("Servo on D2: Min = "));
  Serial.print(parm_value_I);
  Serial.print(F(", Max = "));
  Serial.println(parm_value_M);
  Serial.print(F("Servo on D3: Min = "));
  Serial.print(parm_value_J);
  Serial.print(F(", Max = "));
  Serial.println(parm_value_N);
  Serial.print(F("Servo on D4: Min = "));
  Serial.print(parm_value_K);
  Serial.print(F(", Max = "));
  Serial.println(parm_value_O);
  Serial.print(F("Servo on D5: Min = "));
  Serial.print(parm_value_L);
  Serial.print(F(", Max = "));
  Serial.print(parm_value_P);
}

void setParm(char parm, int value) {
  switch (parm) {
    case 'A':
      parm_value_A = value;
      break;
    case 'D':
      parm_value_D = value;
      break;
    case 'I':
      parm_value_I = value;
      break;
    case 'J':
      parm_value_J = value;
      break;
    case 'K':
      parm_value_K = value;
      break;
    case 'L':
      parm_value_L = value;
      break;
    case 'M':
      parm_value_M = value;
      break;
    case 'N':
      parm_value_N = value;
      break;
    case 'O':
      parm_value_O = value;
      break;
    case 'P':
      parm_value_P = value;
      break;
    case 'V':
      parm_value_V = value;
      break;
    case 'W':
      parm_value_W = value;
      break;
    case 'X':
      parm_value_X = value;
      break;
    case 'Y':
      parm_value_Y = value;
      break;
    case 'Z':
      parm_value_Z = value;
      break;
  }
}

int getParm(char parm) {
  switch (parm) {
    case 'A':
      return parm_value_A;
      break;
    case 'D':
      return parm_value_D;
      break;
    case 'I':
      return parm_value_I;
      break;
    case 'J':
      return parm_value_J;
      break;
    case 'K':
      return parm_value_K;
      break;
    case 'L':
      return parm_value_L;
      break;
    case 'M':
      return parm_value_M;
      break;
    case 'N':
      return parm_value_N;
      break;
    case 'O':
      return parm_value_O;
      break;
    case 'P':
      return parm_value_P;
      break;
    case 'V':
      return parm_value_V;
      break;
    case 'W':
      return parm_value_W;
      break;
    case 'X':
      return parm_value_X;
      break;
    case 'Y':
      return parm_value_Y;
      break;
    case 'Z':
      return parm_value_Z;
      break;
  }
}

void saveRecent(String this_command) {
  bool isDuplicate = 0;
  for (h_idx = 0; h_idx < HST_ELEMENTS; h_idx++) {
    myStr = "";
    myStr.concat(cmdHistory[h_idx]);
    if (this_command.equals(myStr)) {
      isDuplicate = 1;
      h_idx = HST_ELEMENTS;
    }
  }
  if (isDuplicate < 1) {
    for (h_idx = 0; h_idx < HST_ELEMENTS - 1; h_idx++) {
      myStr = "";
      myStr.concat(cmdHistory[h_idx + 1]);
      myStr.toCharArray(cmdHistory[h_idx], HST_SIZE);
    }
    this_command.toCharArray(cmdHistory[HST_ELEMENTS - 1], HST_SIZE);
  }
}

void VaryVerbosity() {
  verbosity_level++;
  if (verbosity_level > 4) {
    verbosity_level = 0;
  }
  Serial.println();
  Serial.print(F("T = "));
  Serial.print(verbosity_level);
}

void sendToServo(String servo_command) {
  if (verbosity_level > 0) {
    Serial.println();
    Serial.print(F("To Servo: "));
    Serial.print(servo_command);
  }
  comm_activity = micros();
  if (servo_command.equals(F("S1990"))) {
    servo_cmd_mode = 1;
  }
  if (servo_command.equals(F("S1989"))) {
    servo_cmd_mode = 1;
  }
  gtSerial2.println();
  last_command = servo_command;
  for (int i = 0; i < servo_command.length(); i++) {
    gtSerial2.print(servo_command.charAt(i));
  }
  gtSerial2.println();
}

void sendToTextSource(String ts_content) {
  if (verbosity_level > 2) {
    Serial.println();
    Serial.print(F("To Text Source: "));
    Serial.print(ts_content);
  }
  comm_activity = micros();
  command_input_enabled = 0;
  transmitText(ts_content);
  transmitEndOfLine_1();
}

void transmitText(String string_content) {
  // always start with a period of time during which the output is HIGH
  // this needs to be long enough for the console to print or
  // otherwise process the message sent before this one
  delayMicroseconds(START_MSG_PERIOD);
  for (int i = 0; i < string_content.length(); i++) {
    transmitChar_1(string_content.charAt(i));
  }
  begin_high = micros();
  BB_input_mode = 1;
}

void transmitNumber(int value) {
  myStr = String(value);
  for (int i = 0; i < myStr.length(); i++) {
    transmitChar_1(myStr.charAt(i));
    Serial.print(myStr.charAt(i));
  }
}

void transmitEndOfLine_1() {
  transmitChar_1(13); // ASCII CR
}

void transmitEndOfTransmission() {
  transmitChar_1(4); // ASCII EOT
  Serial.println();
}

void transmitChar_1(int value) {
  // output to BB_TX_PIN_1
  // seven bits make an ASCII character
  // the bits of the character are always followed by a period during which the output is HIGH
  // this is to allow the receiver to store the character in the message array
  // the text is transmitted without pause other than for that slight delay
  // for consistent timing the pulse durations are in eight simple variables
  // the receiving node is completely dedicated to saving the message chars
  // short period means bit is a zero
  // long period means bit is a one
  // 1 2 3 4 5 6 7 END
  // 3+3+3+3+3+3+3+30   shortest character 0x00 51 microseconds
  // 8+8+8+8+8+8+8+30   longest character  0x7F 86 microseconds
  // min throughput of 11627 characters per second
  // max throughput of 19607 characters per second
  if (value > 127) {
    value = 127;
  }
  if (value > 63) {
    value -= 64;
    tx_period_1 = LONG_PERIOD;
  } else {
    tx_period_1 = SHORT_PERIOD;
  }
  if (value > 31) {
    value -= 32;
    tx_period_2 = LONG_PERIOD;
  } else {
    tx_period_2 = SHORT_PERIOD;
  }
  if (value > 15) {
    value -= 16;
    tx_period_3 = LONG_PERIOD;
  } else {
    tx_period_3 = SHORT_PERIOD;
  }
  if (value > 7) {
    value -= 8;
    tx_period_4 = LONG_PERIOD;
  } else {
    tx_period_4 = SHORT_PERIOD;
  }
  if (value > 3) {
    value -= 4;
    tx_period_5 = LONG_PERIOD;
  } else {
    tx_period_5 = SHORT_PERIOD;
  }
  if (value > 1) {
    value -= 2;
    tx_period_6 = LONG_PERIOD;
  } else {
    tx_period_6 = SHORT_PERIOD;
  }
  if (value > 0) {
    tx_period_7 = LONG_PERIOD;
  } else {
    tx_period_7 = SHORT_PERIOD;
  }
  delayMicroseconds(START_CHAR_PERIOD);
  digitalWrite(BB_TX_PIN_1, LOW);
  delayMicroseconds(tx_period_1);
  digitalWrite(BB_TX_PIN_1, HIGH);
  delayMicroseconds(tx_period_2);
  digitalWrite(BB_TX_PIN_1, LOW);
  delayMicroseconds(tx_period_3);
  digitalWrite(BB_TX_PIN_1, HIGH);
  delayMicroseconds(tx_period_4);
  digitalWrite(BB_TX_PIN_1, LOW);
  delayMicroseconds(tx_period_5);
  digitalWrite(BB_TX_PIN_1, HIGH);
  delayMicroseconds(tx_period_6);
  digitalWrite(BB_TX_PIN_1, LOW);
  delayMicroseconds(tx_period_7);
  digitalWrite(BB_TX_PIN_1, HIGH);
}

String fiveDigitParm(int value) {
  String retString = "";
  if (value < 0) {
    value = 0;
  }
  if (value < 10000) {
    retString += F("0");
  }
  if (value < 1000) {
    retString += F("0");
  }
  if (value < 100) {
    retString += F("0");
  }
  if (value < 10) {
    retString += F("0");
  }
  retString += value;
  return retString;
}

String twoDigitParm(int value) {
  String retString = "";
  if (value < 10) {
    retString += F("0");
  }
  retString += value;
  return retString;
}
